
BLACK = "1;30"
RED = "1;31"
GREEN = "1;32"
YELLOW = "1;33"
BLUE = "1;34"
MAGENTA = "1;35"
CYAN = "1;36"
WHITE = "1;37"


# File display.py
WARNINGFILE = "\n\033[{}m[-]Error: File does not exist.".format("0;31")

# File transform.py
WARNINGCROPPING = "\n0\33[{}m[-]Error: Impossible to crop image.".format("0;31")

# File filter.py
WARNINGIMAGESIZE ="\n\033[{}m[-]Warning: Unknown size of image.".format("0;33")
WARNINGPADDING = "\n\033[{}m[-]Warning: Unknown type of padding.".format("0;33")
WARNINGFILTER = "\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33")

# FIle FT.py
WARNINGFTFILTER = "\n\033[{}m[-]Warning: Unknown type of filter.".format("0;33")






